class FeetechFrame():
    def __init__(self, timestamp, qpos, velocity, control_force):
        self.timestamp = timestamp
        self.qpos = qpos
        self.velocity = velocity
        self.control_force = control_force

    def __repr__(self):
        return f"FeetechFrame(timestamp={self.timestamp}, qpos={self.qpos}, velocity={self.velocity}, control_force={self.control_force})"


import numpy as np
import torch
import time
import genesis as gs

from slobot.genesis import Genesis
from slobot.configuration import Configuration
from slobot.simulation_frame import SimulationFrame

class SoArm100():
    JAW_LENGTH = 0.107
    JAW_WIDTH = 0.008
    JAW_DEPTH = 0.0005

    # Mujoco home position
    HOME_QPOS = [0, -np.pi/2, np.pi/2, np.pi/2, -np.pi/2, 0]

    def sim_qpos(target_qpos):
        mjcf_path = Configuration.MJCF_CONFIG
        arm = SoArm100(mjcf_path=mjcf_path)
        arm.genesis.entity.set_qpos(target_qpos)
        arm.genesis.hold_entity()

    def __init__(self, **kwargs):
        self.step_handler = kwargs.get('step_handler', None)
        # overwrite step handler to delegate to this class first
        kwargs['step_handler'] = self

        self.feetech = kwargs.get('feetech', None)

        self.genesis = Genesis(**kwargs)

        self.camera = self.genesis.camera
        self.entity = self.genesis.entity
        self.fixed_jaw = self.genesis.fixed_jaw

        self.rgb = kwargs.get('rgb', False)
        self.depth = kwargs.get('depth', False)
        self.segmentation = kwargs.get('segmentation', False)
        self.normal = kwargs.get('normal', False)

    def elemental_rotations(self):
        self.go_home()
        pos = self.fixed_jaw.get_pos()
        quat = self.fixed_jaw.get_quat()

        print("pos=", pos)
        print("quat=", quat)

        euler = self.genesis.quat_to_euler(quat)

        print("euler=", euler)

        steps = 2

        # turn the fixed jaw around the global x axis, from vertical to horizontal
        for roll in np.linspace(np.pi/2, 0, steps):
            euler[0] = roll
            quat = self.genesis.euler_to_quat(euler)
            self.genesis.move(self.fixed_jaw, pos, quat)

        # turn the fixed jaw around the global y axis
        for pitch in np.linspace(0, np.pi, steps):
            euler[1] = pitch
            quat = self.genesis.euler_to_quat(euler)
            self.genesis.move(self.fixed_jaw, pos, quat)

        # turn the fixed jaw around the global z axis
        pos = None
        for yaw in np.linspace(0, np.pi/2, steps):
            euler[2] = yaw
            quat = self.genesis.euler_to_quat(euler)
            self.genesis.move(self.fixed_jaw, pos, quat)

    def diff_ik(self):
        center = torch.tensor([0, -0.1, 0.3])
        r = 0.1
        for i in range(0, 1000):
            target_pos = center + torch.tensor([np.cos(i / 360 * np.pi), np.sin(i / 360 * np.pi), 0]) * r

            target_qpos = self.entity.inverse_kinematics(
                link     = self.fixed_jaw,
                pos      = target_pos,
                quat     = None,
            )

            self.entity.control_dofs_position(target_qpos)
            self.genesis.step()

    def lift_fixed_jaw(self):
        qpos_target = Configuration.QPOS_MAP['rotated']
        self.entity.control_dofs_position(qpos_target)

        for i in range(0, 100):
            self.genesis.step()

        print("qpos rotated=", self.entity.get_qpos())

        current_pos = self.fixed_jaw.get_pos()
        print("ee rotated", current_pos)
        current_pos[2] += 0.1

        self.genesis.move(self.fixed_jaw, current_pos, None)

        print("qpos lifted=", self.entity.get_qpos())
        current_pos = self.fixed_jaw.get_pos()
        print("ee lifted", current_pos)

    def stop(self):
        #self.camera.stop_recording(save_to_filename='so_arm_100.mp4')
        self.genesis.stop()

    def go_home(self):
        target_qpos = torch.tensor(SoArm100.HOME_QPOS)
        self.genesis.follow_path(target_qpos)

    def draw_fixed_jaw_arrow(self):
        t = [self.JAW_WIDTH, -self.JAW_LENGTH, -self.JAW_DEPTH]
        self.genesis.draw_arrow(self.fixed_jaw, t)

    def open_jaw(self):
        self.genesis.update_qpos(self.jaw, np.pi/2)

    def handle_step(self) -> SimulationFrame:
        if self.step_handler is None:
            return

        simulation_frame = self.create_simulation_frame()
        self.step_handler.handle_step(simulation_frame)
        return simulation_frame

    def create_simulation_frame(self) -> SimulationFrame:
        current_time = time.time()

        # convert torch tensor to a JSON serializable object
        qpos = self.entity.get_qpos().tolist()
        velocity = self.entity.get_dofs_velocity().tolist()
        force = self.entity.get_dofs_force().tolist()
        control_force = self.entity.get_dofs_control_force().tolist()

        simulation_frame = SimulationFrame(
            timestamp=current_time,
            qpos=qpos,
            velocity=velocity,
            force=force,
            control_force=control_force,
        )

        if self.rgb or self.depth or self.segmentation or self.normal:
            frame = self.camera.render(rgb=self.rgb, depth=self.depth, segmentation=self.segmentation, colorize_seg=True, normal=self.normal)
            rbg_arr, depth_arr, seg_arr, normal_arr = frame
            simulation_frame.rgb = rbg_arr
            simulation_frame.depth = depth_arr
            simulation_frame.segmentation = seg_arr
            simulation_frame.normal = normal_arr

        if self.feetech is not None:
            simulation_frame.feetech_frame = self.feetech.create_feetech_frame()

        return simulation_frame

    def handle_qpos(self, feetech_frame):
        self.entity.set_qpos(feetech_frame.qpos)
        self.genesis.step()


from gradio_client import Client
from slobot.simulation_frame import SimulationFrame
from slobot.configuration import Configuration

import time

class SimClient():
    def __init__(self, **kwargs):
        self.logger = Configuration.logger(__name__)
        url = kwargs['url']
        self.client = Client(url)
        self.step_handler = kwargs['step_handler']

    def run(self, fps):
        job = self.client.submit(fps=fps, api_name="/sim_qpos")
        previous_time = time.time()
        period = 1.0 / fps
        for qpos in job:
            self.logger.info(f"Received qpos {qpos}")
            simulation_frame = SimulationFrame(0, qpos)

            current_time = time.time()
            delta = current_time - (previous_time + period)
            if delta < 0:
                time.sleep(-delta)

            self.step_handler.handle_step(simulation_frame)
            previous_time = max(current_time, previous_time + period)



import time
import gradio as gr
from slobot.image_streams import ImageStreams
from slobot.configuration import Configuration

class GradioQposApp():
    def __init__(self):
        self.image_streams = ImageStreams()
        self.logger = Configuration.logger(__name__)

    def launch(self):
        with gr.Blocks() as demo:
            with gr.Row():
                button = gr.Button()
                fps = gr.Slider(label='FPS', minimum=1, maximum=24, value=3, step=1)
            with gr.Row():
                shoulder_pan = gr.Number(label="shoulder_pan", precision=2)
                shoulder_lift = gr.Number(label="shoulder_lift", precision=2)
                elbow_flex = gr.Number(label="elbow_flex", precision=2)
                wrist_flex = gr.Number(label="wrist_flex", precision=2)
                wrist_roll = gr.Number(label="wrist_roll", precision=2)
                gripper = gr.Number(label="gripper", precision=2)

            button.click(self.sim_qpos, [fps], [shoulder_pan, shoulder_lift, elbow_flex, wrist_flex, wrist_roll, gripper])

        demo.launch()

    def sim_qpos(self, fps):
        res = Configuration.LD
        #sleep_period = 1.0 / fps
        for simulation_frame_paths in self.image_streams.simulation_frame_paths(res, fps, rgb=False, depth=False, segmentation=False, normal=False):
            #time.sleep(sleep_period)
            self.logger.debug(f"Sending qpos {simulation_frame_paths.simulation_frame.qpos}")
            yield simulation_frame_paths.simulation_frame.qpos



import time
import os
import threading
import queue

from PIL import Image

from slobot.configuration import Configuration
from slobot.so_arm_100 import SoArm100
from slobot.video_streams import VideoStreams
from slobot.simulation_frame import SimulationFrame
from slobot.simulation_frame_paths import SimulationFramePaths

# Generate a stream of images from the simulation
class ImageStreams:

    def __init__(self):
        os.makedirs(Configuration.WORK_DIR, exist_ok=True)
        self.queue = queue.Queue()

    def simulation_frame_paths(self, res, fps, rgb=True, depth=False, segmentation=False, normal=False):
        thread = threading.Thread(target=self.run_simulation, args=(res, fps, rgb, depth, segmentation, normal))
        thread.start()

        while True:
            simulation_frame_paths = self.queue.get()
            if simulation_frame_paths is None:
                break

            yield simulation_frame_paths

        thread.join()

    def run_simulation(self, res, fps, rgb=True, depth=False, segmentation=False, normal=False):
        self.start(res, fps, rgb=rgb, depth=depth, segmentation=segmentation, normal=normal)

        mjcf_path = Configuration.MJCF_CONFIG
        arm = SoArm100(mjcf_path=mjcf_path, step_handler=self, res=res, fps=fps, show_viewer=False, rgb=rgb, depth=depth, segmentation=segmentation, normal=normal)
        arm.elemental_rotations()
        arm.stop()

        self.queue.put(None) # add poison pill

    def start(
        self,
        res,
        fps,
        rgb=True,
        depth=False,
        segmentation=False,
        normal=False
    ):
        self.res = res
        self.fps = fps
        self.segment_id = 0

        self.frame_enabled = [rgb, depth, segmentation, normal]

    def handle_step(self, simulation_frame: SimulationFrame):
        if simulation_frame.depth is not None:
           # colorize depth
           simulation_frame.depth = VideoStreams.logarithmic_depth_to_rgb(simulation_frame.depth)

        simulation_frame_paths = self.transcode_frame(simulation_frame)
        self.queue.put(simulation_frame_paths)

    def transcode_frame(self, simulation_frame: SimulationFrame) -> SimulationFramePaths:
        date_time = time.strftime('%Y%m%d_%H%M%S')

        simulation_frame_images = []
        for frame_id in range(len(VideoStreams.FRAME_TYPES)):
            if not self.frame_enabled[frame_id]:
                continue

            filename = self._filename(VideoStreams.FRAME_TYPES[frame_id], date_time, self.segment_id)
            self.create_image_paths(simulation_frame.frame(frame_id), filename)
            simulation_frame_images.append(filename)

        self.segment_id += 1

        return SimulationFramePaths(simulation_frame=simulation_frame, paths=simulation_frame_images)

    def create_image_paths(self, typed_array, filename):
        image = Image.fromarray(typed_array, mode='RGB')
        image.save(filename)
        return filename

    def _filename(self, frame_type, date_time, segment_id):
        return f"{Configuration.WORK_DIR}/{frame_type}_{date_time}_{segment_id}.webp"

